#version 150
/// BEGIN AUTOGENERATED HEADER
#define CONC(a,b) a##b

#define IF(c, t, e) CONC(IF_, c)(t, e)

#define IF_UNDEF(t, e) e
#define IF_0(t, e) t
#define IF_1(t, e) t
#define IF_2(t, e) t
#define IF_3(t, e) t
#define IF_4(t, e) t
#define IF_5(t, e) t
#define IF_6(t, e) t
#define IF_7(t, e) t
#define IF_8(t, e) t
#define IF_9(t, e) t
#define IF_10(t, e) t
#define IF_11(t, e) t
#define IF_12(t, e) t
#define IF_13(t, e) t
#define IF_14(t, e) t
#define IF_15(t, e) t
#define IF_16(t, e) t
// Workaround for intel HD:
#define IF_00(t, e) t
#ifdef GL_SPIRV
	#define gl_VertexID gl_VertexIndex

	#define LAYOUT_SAMPLER2(s, b) layout(set = s, binding = b)
	#define LAYOUT_SAMPLER(s, b) layout(set = s, binding = b)
	#define LAYOUT_UBO(s, b) layout(std140, set = s, binding = b)
	#define LAYOUT_PUSH_CONSTANT() layout(push_constant, std140)

	#define VERTEX_INPUT(loc, type, name, defaultValue) IF(loc, layout (location = loc) in type name, const type name = defaultValue)
		
	#define INPUT_MOD(loc, mod, type, name, defaultValue) IF(loc, layout (location = loc) mod in type name, const type name = defaultValue)
		
	#define INPUT(loc, type, name, defaultValue) IF(loc, layout (location = loc) in type name, const type name = defaultValue)

	#define OUTPUT(loc, type, name) IF(loc, layout (location = loc) out type name, type name)
	
	#define OUTPUT_MOD(loc, mod, type, name) IF(loc, layout (location = loc) mod out type name, type name)
	
	#define WINDING ccw

	#define LOCATION(loc) layout(location = loc) 
#else
	#extension GL_ARB_separate_shader_objects : require
	#extension GL_ARB_explicit_attrib_location : require
	#extension GL_ARB_shading_language_420pack : require

#ifdef USE_SAMPLER2
	#define LAYOUT_SAMPLER2(s, b)
#else
	#define LAYOUT_SAMPLER2(s, b) layout(binding = b)
#endif

	#define LAYOUT_SAMPLER(s, b) layout(binding = b)
	#define LAYOUT_UBO(s, b) layout(std140, binding = b)
	#define LAYOUT_PUSH_CONSTANT() layout(std140, binding = 2)

	#define VERTEX_INPUT(loc, type, name, defaultValue) IF(loc, layout (location = loc) in type name, const type name = defaultValue)
		
	#define INPUT_MOD(loc, mod, type, name, defaultValue) mod in type name
		
	#define INPUT(loc, type, name, defaultValue) in type name

	#define OUTPUT(loc, type, name) out type name
	
	#define OUTPUT_MOD(loc, mod, type, name) mod out type name
	
	#define WINDING ccw

	#define LOCATION(loc)
#endif
/// END AUTOGENERATED HEADER

LAYOUT_SAMPLER(2, 25) uniform sampler2D materialIndexTex;

LAYOUT_UBO(1, 11) uniform Terrain {
	vec2 terrainStart;
	vec2 terrainMinMax;
	
	vec4 terrainTileData1; // tileSize, matBorder, ambSize, ambBorder
	float terrainTileData2; // hmBorder
} u_terrain;

LAYOUT_UBO(2, 14) uniform Tile {
	ivec4 materialMap[17 * 16];
} u_tile; // TODO

LAYOUT_PUSH_CONSTANT() uniform PushConstant {
	vec4 blockCoordDim;
	vec2 tileSubPos;
	vec2 tileOffset;
	float distanceScaling;
	float alphaThreshold;
	int numSamples;
	vec4 color;
} u_push; // TODO

LOCATION(0) out VertexData { // TODO: pack
	vec2 texCoord;
	int matType;
	int sampleIndex;
} outData;

#define TILE_GROUPS 4

int mapMaterial(int i, int offset) {
	return u_tile.materialMap[i + offset + 1].x - 1;
}

int getNumMaterials(int offset) {
	return u_tile.materialMap[offset].x;
}

int getOffset(vec2 tileSubPos) {
	return 17 * (4 * int(tileSubPos.x) + int(tileSubPos.y));
}

const int sampleSize = 16;
const vec2[sampleSize] sampleCoords = vec2[sampleSize](
	vec2(1, 1),
	vec2(2, 3),
	vec2(3, 0),
	vec2(0, 2),
	
	vec2(3, 2),
	vec2(0, 0),
	vec2(1, 3),
	vec2(2, 1),
	
	vec2(0, 3),
	vec2(3, 1),
	vec2(1, 0),
	vec2(2, 2),

	vec2(0, 1),
	vec2(3, 3),
	vec2(2, 0),
	vec2(1, 2)
);

const float[sampleSize] offsets = float[sampleSize](
	0.0, 0.5, 0.125, 0.625,
	0.75, 0.25, 0.875, 0.375,
	0.1875, 0.6875, 0.0625, 0.5625,
	0.9375, 0.4375, 0.8125, 0.3125
);

vec2 addBorder(vec2 coord, float res, float border) {
	float actualRes = res * TILE_GROUPS + 2.0 * border;
 	return ((coord + u_push.tileSubPos) * res + border) / actualRes;
}

ivec2 addBorderN(vec2 coord, float res, float border) {
	float actualRes = res + 2.0 * border;
   	vec2 pos2 = ((coord * res + border) / actualRes + u_push.tileSubPos); 

	int actualResInt = int(actualRes);
	ivec2 of = actualResInt * ivec2(u_push.tileSubPos);
	return clamp(ivec2(pos2 * actualResInt), of, actualResInt + of - 1);
}

int safeToInt(float val) {
	return int(val * 255.0 + 0.5);
}

void main() {
	outData.sampleIndex = gl_VertexID % u_push.numSamples;
	int cellIndex = gl_VertexID / u_push.numSamples;
	int ix = cellIndex % int(u_push.blockCoordDim.z);
	int iy = cellIndex / int(u_push.blockCoordDim.z);
	vec2 cellCoord = vec2(ix, iy); // meters
	
	vec2 sampleCoord = sampleCoords[(outData.sampleIndex + 3 * ix * 7 * iy) % sampleSize] / 4; // sample offset
	sampleCoord += vec2(offsets[iy % sampleSize], offsets[ix % sampleSize]) / 4 + 1.0 / 32; // random offset, meters
	
	vec2 rpos = u_push.blockCoordDim.xy + (cellCoord + sampleCoord);
	
	outData.texCoord = rpos / u_terrain.terrainTileData1.x;

	outData.matType = safeToInt(texelFetch(materialIndexTex, addBorderN(outData.texCoord, u_terrain.terrainTileData1.x, u_terrain.terrainTileData1.y), 0).r);

	int offset = getOffset(u_push.tileSubPos);

	if (getNumMaterials(offset + 1) != -1) outData.matType = mapMaterial(outData.matType, offset) + 1;
}
