#version 150
/// BEGIN AUTOGENERATED HEADER
#define CONC(a,b) a##b

#define IF(c, t, e) CONC(IF_, c)(t, e)

#define IF_UNDEF(t, e) e
#define IF_0(t, e) t
#define IF_1(t, e) t
#define IF_2(t, e) t
#define IF_3(t, e) t
#define IF_4(t, e) t
#define IF_5(t, e) t
#define IF_6(t, e) t
#define IF_7(t, e) t
#define IF_8(t, e) t
#define IF_9(t, e) t
#define IF_10(t, e) t
#define IF_11(t, e) t
#define IF_12(t, e) t
#define IF_13(t, e) t
#define IF_14(t, e) t
#define IF_15(t, e) t
#define IF_16(t, e) t
#ifdef GL_SPIRV
	#define LAYOUT_SAMPLER(s, b) layout(set = s, binding = b)
	#define LAYOUT_UBO(s, b) layout(std140, set = s, binding = b)
	#define LAYOUT_PUSH_CONSTANT() layout(push_constant, std140)

	#define VERTEX_INPUT(loc, type, name, defaultValue) IF(loc, layout (location = loc) in type name, const type name = defaultValue)
		
	#define INPUT_MOD(loc, mod, type, name, defaultValue) IF(loc, layout (location = loc) mod in type name, const type name = defaultValue)
		
	#define INPUT(loc, type, name, defaultValue) IF(loc, layout (location = loc) in type name, const type name = defaultValue)

	#define OUTPUT(loc, type, name) IF(loc, layout (location = loc) out type name, type name)
	
	#define OUTPUT_MOD(loc, mod, type, name) IF(loc, layout (location = loc) mod out type name, type name)
	
	#define WINDING ccw

	#define LOCATION(loc) layout(location = loc) 
#else
	#extension GL_ARB_separate_shader_objects : require
	#extension GL_ARB_explicit_attrib_location : require
	#extension GL_ARB_shading_language_420pack : require

	#define LAYOUT_SAMPLER(s, b) layout(binding = b)
	#define LAYOUT_UBO(s, b) layout(std140, binding = b)
	#define LAYOUT_PUSH_CONSTANT() layout(std140, binding = 2)

	#define VERTEX_INPUT(loc, type, name, defaultValue) IF(loc, layout (location = loc) in type name, const type name = defaultValue)
		
	#define INPUT_MOD(loc, mod, type, name, defaultValue) mod in type name
		
	#define INPUT(loc, type, name, defaultValue) in type name

	#define OUTPUT(loc, type, name) out type name
	
	#define OUTPUT_MOD(loc, mod, type, name) mod out type name
	
	#define WINDING ccw

	#define LOCATION(loc)
#endif
/// END AUTOGENERATED HEADER

//#define USE_SHADOW_MAP

LAYOUT_UBO(0, 10) uniform View {
	mat4 projView;
	mat4 projViewInverse;
	mat4 proj;
	mat4 projInverse;
	mat4 view;
	mat4 viewInverse;
	vec3 camPos;
	float lodFactor;
	float near;
	float far;
} u_view;

LAYOUT_UBO(0, 6) uniform Light {
	vec3 lightDir;
	float lightScale;
	mat4 lightTransf;
	mat4 lightInvTransf;
	float ambientScale;
} u_light;

LAYOUT_UBO(0, 5) uniform Atmosphere {
	float fogDensity;
	vec2 fogDist;
	vec3 sunColor;
	vec3 rayleighExtCoeff;
	float rayleighScaleHeight;
	vec3 mieScattCoeff;
	float mieScaleHeight;
	float phaseG;
	float viewSamples;
} u_atmosphere;

#ifdef USE_SHADOW_MAP
LAYOUT_UBO(0, 8)  uniform ShadowMap {
	mat4 lightMat;
} u_shadowMap;
LAYOUT_SAMPLER(0, 19) uniform sampler2DShadow depthTex;
#endif

OUTPUT(VPOS_LOC, vec3, vpos);
OUTPUT(FOGFACT_LOC, float, fogFactor);
OUTPUT(EXTFACT_LOC, vec3, extFactor);
OUTPUT(INSCATT_LOC, vec3, inScatt);

#ifdef USE_SHADOW_MAP
float compShadow(vec3 vpos) {
	vec4 proj = u_shadowMap.lightMat * vec4(vpos, 1.0);
	proj.xy /= proj.w;

	if (any(lessThan(proj.xy, vec2(-1.0, -1.0))) || any(greaterThan(proj.xy, vec2(1.0, 1.0)))) {
		return 1.0;
	}

	proj.xyz = .5 * proj.xyz + vec3(.5);

	return texture(depthTex, proj.xyz);
}
#endif

// TODO change calcFogFactor in fog.gs too!
float calcFogFactor2(vec3 pos) {
	vec3 camPos = vec3(u_view.viewInverse[3]);
	//calcFogFactor(length(pos - camPos));

	vec3 weights = vec3(1.0, 1.0, 1.0);		// 1.0, 1.0, .25
	float dist = length(weights * (pos - camPos));
	//return 1.0 - clamp(exp(-fogDensity * (dist - fogDist.x) / fogDist.y), .0, 1.0);

	vec2 fogDist = u_atmosphere.fogDist * u_view.far;

	float f = clamp((dist - fogDist.x) / fogDist.y, .0, 1.0);
	return u_atmosphere.fogDensity * pow(f, 4.5);
}

float RayleightPhase(float cosTheta2) {
	return .589 * (1 + cosTheta2);
}

float HenyeyGreensteinFn(float g, float cosTheta, float cosTheta2) {
	float g2 = g*g;
	float a = (1.0 - g2) / (2.0 + g2);
	float b = (1.0 + cosTheta2) / pow(1.0 + g2 - 2.0 * g * cosTheta, 3.0 / 2.0);
	return 4 * 3.0 / 25.13 * a * b;
}

void calcFogFactor(vec3 pos) {
	vpos = pos;
	extFactor = vec3(1.0);
	inScatt = vec3(.0);
	fogFactor = calcFogFactor2(pos);
	
	if (u_atmosphere.viewSamples <= .0) return;

	vec3 camPos = vec3(u_view.viewInverse[3]);
	vec3 viewVec = camPos - pos;
	
	if (camPos.z < .0 && pos.z >= .0) {
		float t = -camPos.z / viewVec.z;
		camPos = camPos + t * viewVec;
	}
	
	viewVec = camPos - pos;
	
	float dist = length(viewVec);
	
	vec3 extCoeff = u_atmosphere.rayleighExtCoeff + u_atmosphere.mieScattCoeff / .9;
	
	float h0 = max(camPos.z, .0);
	float h1 = max(pos.z, .0);
	
	float rD = .0;
	float mD = .0;
	
	float step = dist / u_atmosphere.viewSamples;
	
	vec3 rayleightInScatt = vec3(.0);
	vec3 mieInScatt = vec3(.0);
	
	for (int i = 0; i < u_atmosphere.viewSamples; i++) {
		float h = mix(h0, h1, (i + .5) / u_atmosphere.viewSamples);
		float pr = exp(-h / u_atmosphere.rayleighScaleHeight) * step;
		float pm = exp(-h / u_atmosphere.mieScaleHeight) * step;
		
		rD += pr;
		mD += pm;
		
		extFactor = exp(-u_atmosphere.rayleighExtCoeff * rD - u_atmosphere.mieScattCoeff * mD);
		
#ifdef USE_SHADOW_MAP
		if (compShadow(mix(camPos, pos, (i + .5) / u_atmosphere.viewSamples)) == 0) continue;
#endif
	
		rayleightInScatt += extFactor * pr;
		mieInScatt += extFactor * pm;
	}
	
	float cosTheta = clamp(dot(normalize(viewVec), u_light.lightDir), -1.0, 1.0);
	float cosTheta2 = cosTheta*cosTheta;
	
	vec3 rayleightScatt = u_atmosphere.rayleighExtCoeff * RayleightPhase(cosTheta2);
	vec3 mieScatt = u_atmosphere.mieScattCoeff * HenyeyGreensteinFn(u_atmosphere.phaseG, cosTheta, cosTheta2);
	
	inScatt = rayleightInScatt * rayleightScatt + mieInScatt * mieScatt;
}
